@@ -19,7 +19,7 @@ from functools import reduce
 import six
 from decorator import decorator
 
-__version__ = '4.0.0'
+from appoptics_apm.version import __version__
 
 # "invalid name ... For type constant"
 # pylint interprets all module-level variables as being 'constants'.
@@ -36,7 +36,7 @@ OBOE_TRACE_THROUGH = 2
 # oboe modules
 OBOE_MODULE_ALL = -1
 
-# Debug level from liboboe: oboe_debug.h
+# maps Debug level from liboboe: oboe_debug.h
 OBOE_DEBUG_DISABLE = -1
 OBOE_DEBUG_FATAL = 0
 OBOE_DEBUG_ERROR = 1
@@ -46,17 +46,51 @@ OBOE_DEBUG_LOW = 4
 OBOE_DEBUG_MEDIUM = 5
 OBOE_DEBUG_HIGH = 6
 
-# Map the oboe debug level to python logging level
-_logging_map = {
-    OBOE_DEBUG_DISABLE: logging.CRITICAL,
-    OBOE_DEBUG_FATAL: logging.CRITICAL,
-    OBOE_DEBUG_ERROR: logging.ERROR,
-    OBOE_DEBUG_WARNING: logging.WARNING,
-    OBOE_DEBUG_INFO: logging.INFO,
-    OBOE_DEBUG_LOW: logging.DEBUG,
-    OBOE_DEBUG_MEDIUM: logging.DEBUG,
-    OBOE_DEBUG_HIGH: logging.DEBUG
-}
+
+class _OboeDebugLevel(object):
+    debug_levels = {
+        'OBOE_DEBUG_DISABLE': OBOE_DEBUG_DISABLE,
+        'OBOE_DEBUG_FATAL': OBOE_DEBUG_FATAL,
+        'OBOE_DEBUG_ERROR': OBOE_DEBUG_ERROR,
+        'OBOE_DEBUG_WARNING': OBOE_DEBUG_WARNING,
+        'OBOE_DEBUG_INFO': OBOE_DEBUG_INFO,
+        'OBOE_DEBUG_LOW': OBOE_DEBUG_LOW,
+        'OBOE_DEBUG_MEDIUM': OBOE_DEBUG_MEDIUM,
+        'OBOE_DEBUG_HIGH': OBOE_DEBUG_HIGH,
+    }
+    # maps oboe debug level to python logging level
+    logging_map = {
+        OBOE_DEBUG_DISABLE: logging.CRITICAL,
+        OBOE_DEBUG_FATAL: logging.CRITICAL,
+        OBOE_DEBUG_ERROR: logging.ERROR,
+        OBOE_DEBUG_WARNING: logging.WARNING,
+        OBOE_DEBUG_INFO: logging.INFO,
+        OBOE_DEBUG_LOW: logging.DEBUG,
+        OBOE_DEBUG_MEDIUM: logging.DEBUG,
+        OBOE_DEBUG_HIGH: logging.DEBUG
+    }
+
+    def __init__(self, level):
+        """Creates a _OboeDebugLevel instance from provided level argument. If level can not be converted to a valid
+        debug level, it will create a _OboeDebugLevel instance with default_level"""
+        try:
+            level = int(level)
+            self._level = level if level in list(self.debug_levels.values()) else self.default_level()
+        except (ValueError, TypeError):
+            self._level = self.debug_levels[level] if level in self.debug_levels else self.default_level()
+
+    @classmethod
+    def default_level(cls):
+        """Returns integer representation of default debugging level"""
+        return cls.debug_levels['OBOE_DEBUG_WARNING']
+
+    def is_disable(self):
+        return self._level == self.debug_levels['OBOE_DEBUG_DISABLE']
+
+    def python_logging_level(self):
+        """Returns corresponding python logging module level"""
+        return self.logging_map[self._level]
+
 
 # Sample rate configuration sources
 OBOE_SAMPLE_RATE_SOURCE_FILE = 1
@@ -86,7 +120,7 @@ def _apm_str(non_string):
         # 'utf-8 encoding is more safe if there is regex processing etc. for
         # consistence encode unicode using utf-8
         return non_string.encode('utf-8')
-    except:
+    except Exception:
         return non_string
 
 
@@ -102,7 +136,9 @@ def _get_logger():
         logging.NullHandler = NullHandler
 
     _logger = logging.getLogger(__name__)
-    if ('APPOPTICS_DEBUG_LEVEL' in os.environ and int(os.environ['APPOPTICS_DEBUG_LEVEL']) == OBOE_DEBUG_DISABLE):
+    log_level_envv = os.getenv('APPOPTICS_DEBUG_LEVEL', None)
+    log_level = _OboeDebugLevel(log_level_envv)
+    if log_level.is_disable():
         _logger.addHandler(logging.NullHandler())
         _logger.disabled = True
     else:
@@ -110,7 +146,15 @@ def _get_logger():
         f = logging.Formatter('%(asctime)s [ %(name)s %(levelname)-8s p#%(process)d.%(thread)d] %(message)s')
         sh.setFormatter(f)
         _logger.addHandler(sh)  # use sys.stderr; see oboeware #63
-        _logger.setLevel(_logging_map[int(os.getenv('APPOPTICS_DEBUG_LEVEL', str(OBOE_DEBUG_WARNING)))])
+        _logger.setLevel(log_level.python_logging_level())
+        if log_level_envv is not None and log_level_envv not in _OboeDebugLevel.debug_levels:
+            try:
+                if int(str(log_level_envv)) not in _OboeDebugLevel.logging_map:
+                    raise ValueError
+            except ValueError:
+                _logger.warning(
+                    "Misconfigured APPOPTICS_DEBUG_LEVEL ignored. Defaulted to debug_level {}".format(log_level._level))
+
     return _logger
 
 
@@ -288,7 +332,7 @@ class AppOpticsApmConfig(object):
             #'sample_rate' and 'tracing_mode' do not have any default values
             'collector': '',  # the collector address in host:port format.
             'reporter': '',  # the reporter mode, either 'udp' or 'ssl'.
-            'debug_level': OBOE_DEBUG_WARNING,
+            'debug_level': _OboeDebugLevel.default_level(),
             'warn_deprecated': True,
             'service_key': '',
             'hostname_alias': '',
@@ -307,6 +351,7 @@ class AppOpticsApmConfig(object):
             'reporter_file_single': 0,
             'enable_sanitize_sql': True,
             'inst_enabled': defaultdict(lambda: True),
+            'log_trace_id': 'never',
             'transaction': defaultdict(lambda: True),
             'inst': defaultdict(lambda: True)
         }
@@ -383,9 +428,18 @@ class AppOpticsApmConfig(object):
                 self._config[key] = val.lower()
             elif keys == ['debug_level']:
                 val = int(val)
-                if val not in range(-1, 7):
+                if val not in list(_OboeDebugLevel.debug_levels.values()):
                     raise ValueError
                 self._config[key] = val
+            elif keys == ['log_trace_id']:
+                if not isinstance(val, six.string_types) or val.lower() not in [
+                        'never',
+                        'sampled',
+                        'traced',
+                        'always',
+                ]:
+                    raise ValueError
+                self._config[key] = val.lower()
             elif (keys[0] == 'inst' and keys[1] in self.supported_instrumentations and keys[2] == 'collect_backtraces'):
                 self._config[keys[0]][keys[1]][keys[2]] = _convert_to_bool(val)
             elif isinstance(sub_dict, dict) and keys[-1] in sub_dict:
@@ -396,10 +450,10 @@ class AppOpticsApmConfig(object):
                 sub_dict[keys[-1]] = val
             else:
                 logger.warning("Ignore invalid configuration key: {}".format('.'.join(keys)))
-        except ValueError:
+        except (ValueError, TypeError):
             logger.warning(
                 'Ignore config option with invalid (non-convertible or out-of-range) type: ' +
-                '.'.join(keys if keys[0] not in ['inst', 'transaction'] else key[1:]))
+                '.'.join(keys if keys[0] not in ['inst', 'transaction'] else keys[1:]))
 
     def update_with_cnf_file(self, cnf_path):
         """Update the settings with the config file, if any."""
@@ -439,6 +493,11 @@ class AppOpticsApmConfig(object):
             if curr_sample_rate is not None:
                 SwigContext.setDefaultSampleRate(int(curr_sample_rate * 1e6))
 
+        elif key == 'log_trace_id':
+            self._set_config_value(key, value)
+            from appoptics_apm.inst_logging import log_injector
+            log_injector.set_injection_mode(self._config[key])
+
         elif key in ('enable_sanitize_sql', 'warn_deprecated'):
             self._set_config_value(key, value)
         else:
@@ -476,9 +535,9 @@ def appoptics_apm_init():
     if sys_is_traceable():
         # APPOPTICS_COLLECTOR is fetched by underlying library
         # Unfortunately, this function does not return anything for now.
-        oboe_log_level = config['debug_level']
-        logger.setLevel(_logging_map[int(oboe_log_level)])
-        if oboe_log_level == OBOE_DEBUG_DISABLE:
+        oboe_log_level = _OboeDebugLevel(config['debug_level'])
+        logger.setLevel(oboe_log_level.python_logging_level())
+        if oboe_log_level.is_disable():
             disable_logger()
         reporter = _reporter()
         if reporter.init_status == 0:  # OBOE_INIT_OK
@@ -1806,10 +1865,6 @@ def report_layer_init(layer='Python', keys=None):
     ctx.report_status(evt)
 
 
-# Report an status event after everything is done.
-report_layer_init('Python')
-
-
 class oboe_ready_code(object):
     OBOE_SERVER_RESPONSE_UNKNOWN = (0, "Oboe server : unknown error")
     OBOE_SERVER_RESPONSE_OK = (1, "Oboe server : is ready")
@@ -1855,3 +1910,5 @@ def appoptics_ready(wait_milliseconds=3000, integer_response=False):
 
 """The AppOptics APM initializer will be executed when this package is imported by user application"""
 appoptics_apm_init()
+# Report an status event after everything is done.
+report_layer_init('Python')
